import os

from tap_tester import connections, runner
from tap_tester.logger import LOGGER
from base import DixaBaseTest
from datetime import datetime


class DixaStartDateTest(DixaBaseTest):
    start_date = "2022-06-22T00:00:00Z"

    @staticmethod
    def name():
        return "tap_tester_dixa_start_date_test"


    def test_run(self):
        """Instantiate start date according to the desired data set and run the test"""

        self.start_date_1 = self.get_properties().get('start_date')
        self.start_date_2 = self.timedelta_formatted(self.start_date_1, days=5)

        self.start_date = self.start_date_1

        expected_streams = self.expected_streams()

        ##########################################################################
        ### First Sync
        ##########################################################################

        # instantiate connection
        conn_id_1 = connections.ensure_connection(self)

        # run check mode
        found_catalogs_1 = self.run_and_verify_check_mode(conn_id_1)

        # table and field selection
        test_catalogs_1_all_fields = [catalog for catalog in found_catalogs_1
                                      if catalog.get('stream_name') in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id_1, test_catalogs_1_all_fields)

        # run initial sync
        record_count_by_stream_1 = self.run_and_verify_sync(conn_id_1)
        synced_records_1 = runner.get_records_from_target_output()

        ##########################################################################
        ### Update START DATE Between Syncs
        ##########################################################################

        LOGGER.info("REPLICATION START DATE CHANGE: {} ===>>> {} ".format(self.start_date, self.start_date_2))
        self.start_date = self.start_date_2

        ##########################################################################
        ### Second Sync
        ##########################################################################

        # create a new connection with the new start_date
        conn_id_2 = connections.ensure_connection(self, original_properties=False)

        # run check mode
        found_catalogs_2 = self.run_and_verify_check_mode(conn_id_2)

        # table and field selection
        test_catalogs_2_all_fields = [catalog for catalog in found_catalogs_2
                                      if catalog.get('stream_name') in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id_2, test_catalogs_2_all_fields)

        # run sync
        record_count_by_stream_2 = self.run_and_verify_sync(conn_id_2)
        synced_records_2 = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_primary_keys = self.expected_primary_keys()[stream]
                expected_start_date_1 = self.start_date_1
                expected_start_date_2 = self.start_date_2

                # collect information for assertions from syncs 1 & 2 base on expected values
                record_count_sync_1 = record_count_by_stream_1.get(stream, 0)
                record_count_sync_2 = record_count_by_stream_2.get(stream, 0)
                primary_keys_list_1 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_1.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_list_2 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_2.get(stream).get('messages')
                                       if message.get('action') == 'upsert']

                # Verify there are no duplicate records in replicated records
                self.assertEquals(len(primary_keys_list_1), len(set(primary_keys_list_1)))
                self.assertEquals(len(primary_keys_list_2), len(set(primary_keys_list_2)))

                primary_keys_sync_1 = set(primary_keys_list_1)
                primary_keys_sync_2 = set(primary_keys_list_2)

                # collect information specific to incremental streams from syncs 1 & 2
                expected_replication_key = next(iter(self.expected_replication_keys().get(stream)))
                replication_dates_1 =[row.get('data').get(expected_replication_key) for row in
                                      synced_records_1.get(stream, {'messages': []}).get('messages', [])
                                      if row.get('data')]
                replication_dates_2 =[row.get('data').get(expected_replication_key) for row in
                                      synced_records_2.get(stream, {'messages': []}).get('messages', [])
                                      if row.get('data')]

                # # Verify replication key is greater or equal to start_date for sync 1
                for replication_date in replication_dates_1:
                    if isinstance(replication_date, int):
                        expected_start_date_1 = self.date_value_to_epoch(expected_start_date_1)
                        expected_start_date_2 = self.date_value_to_epoch(expected_start_date_2)

                    self.assertGreaterEqual(
                        replication_date, expected_start_date_1,
                            msg="Report pertains to a date prior to our start date.\n" +
                            "Sync start_date: {}\n".format(expected_start_date_1) +
                            "Record date: {} ".format(replication_date)
                    )

                # Verify replication key is greater or equal to start_date for sync 2
                for replication_date in replication_dates_2:
                    self.assertGreaterEqual(
                        replication_date, expected_start_date_2,
                            msg="Report pertains to a date prior to our start date.\n" +
                            "Sync start_date: {}\n".format(expected_start_date_2) +
                            "Record date: {} ".format(replication_date)
                    )

                # Verify the number of records replicated in sync 1 is greater than the number
                # of records replicated in sync 2
                self.assertGreater(record_count_sync_1, record_count_sync_2)

                # Verify the records replicated in sync 2 were also replicated in sync 1
                self.assertTrue(primary_keys_sync_2.issubset(primary_keys_sync_1))



from datetime import datetime

from tap_tester import connections, menagerie, runner
from base import DixaBaseTest
from tap_tester.logger import LOGGER


class DixaInterruptedSyncTest(DixaBaseTest):
    """Test tap sets a bookmark and respects it for the next sync of a stream"""

    def name(self):
        return "tap_tester_dixa_bookmark_test"

    def test_run(self):
        """
        - Verify an interrupted sync can resume based on the currently_syncing and stream level bookmark value.
        - Verify only records with replication-key values greater than or equal to the stream level bookmark 
          are replicated on the resuming sync for the interrupted stream.
        - Verify the yet-to-be-synced streams are replicated following the interrupted stream in the resuming sync. 
          (All yet-to-be-synced streams must replicate before streams that were already synced)
        """
        
        expected_streams = self.expected_streams()
        expected_replication_keys = self.expected_replication_keys()
        expected_replication_methods = self.expected_replication_method()

        ##########################################################################
        # First Sync
        ##########################################################################
        conn_id = connections.ensure_connection(self)

        # Run in check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        catalog_entries = [catalog for catalog in found_catalogs
                           if catalog.get('tap_stream_id') in expected_streams]
        
        self.perform_and_verify_table_and_field_selection(
            conn_id, catalog_entries)

        # Run a first sync job using orchestrator
        first_sync_record_count = self.run_and_verify_sync(conn_id)
        first_sync_records = runner.get_records_from_target_output()
        first_sync_bookmarks = menagerie.get_state(conn_id)

        ##########################################################################
        # Update State Between Syncs
        ##########################################################################
        
        LOGGER.info("Current Bookmark: {}".format(first_sync_bookmarks))
        interrupted_sync_states = { 'currently_syncing': 'messages',
                                    'bookmarks': {  'conversations': first_sync_bookmarks.get('bookmarks', {}).get('conversations', {}),
                                                    'messages': {list(expected_replication_keys['messages'])[0]: 1655856000000},
                                                    'activity_logs': {list(expected_replication_keys['activity_logs'])[0]: self.start_date}}}        
        completed_streams = {'conversations'}
        pending_streams = {'activity_logs'}

        menagerie.set_state(conn_id, interrupted_sync_states)

        ##########################################################################
        # Second Sync
        ##########################################################################

        second_sync_record_count = self.run_and_verify_sync(conn_id)
        second_sync_records = runner.get_records_from_target_output()
        second_sync_bookmarks = menagerie.get_state(conn_id)

        ##########################################################################
        # Test By Stream
        ##########################################################################

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_replication_method = expected_replication_methods[stream]

                # collect information for assertions from syncs 1 & 2 base on expected values
                first_sync_count = first_sync_record_count.get(stream, 0)
                second_sync_count = second_sync_record_count.get(stream, 0)
                first_sync_messages = [record.get('data') for record in
                                       first_sync_records.get(
                                           stream, {}).get('messages', [])
                                       if record.get('action') == 'upsert']
                second_sync_messages = [record.get('data') for record in
                                        second_sync_records.get(
                                            stream, {}).get('messages', [])
                                        if record.get('action') == 'upsert']
                first_bookmark_value = first_sync_bookmarks.get('bookmarks', {stream: None}).get(stream)
                second_bookmark_value = second_sync_bookmarks.get('bookmarks', {stream: None}).get(stream)

                if expected_replication_method == self.INCREMENTAL:

                    # collect information specific to incremental streams from syncs 1 & 2
                    replication_key = next(iter(expected_replication_keys[stream]))
                    
                    interrupted_bookmark_value = interrupted_sync_states['bookmarks'][stream]

                    if stream in completed_streams:
                        # Verify at least 1 record was replicated in the third sync
                        if second_bookmark_value[replication_key] == first_bookmark_value[replication_key]:
                            self.assertEquals(second_sync_count,
                                            1, 
                                            msg="Incorrent bookmarking for {}, at least one record should be replicated".format(stream))
                        else:
                            self.assertGreater(second_sync_count,
                                                1,
                                                msg="Incorrent bookmarking for {}, more than one records should be replicated if second sync bookmark is greater than first sync".format(stream))

                    elif stream == interrupted_sync_states.get('currently_syncing', None):
                        # For interrupted stream records sync count should be less equals
                        self.assertLessEqual(second_sync_count,
                                            first_sync_count,
                                            msg="For interrupted stream, seconds sync record count should be lesser or equal to first sync".format(stream))

                    elif stream in pending_streams:
                        # First sync and second sync record count match
                        if second_bookmark_value[replication_key] == first_bookmark_value[replication_key]:
                            self.assertEquals(second_sync_count,
                                            first_sync_count,
                                            msg="For pending sync stream, if bookmark values are same for first and second sync, record should match".format(stream))
                        else:
                            self.assertGreaterEqual(second_sync_count,
                                                    first_sync_count,
                                                    msg="For pending sync streams, second sync record count should be more than first sync".format(stream))     

                    else:
                        raise Exception("Invalid state of stream {0} in interrupted state, please update appropriate state for the stream".format(stream))
                    
                    for record in second_sync_messages:
                        # Verify the second sync replication key value is Greater or Equal to the first sync bookmark
                        replication_key_value = record.get(replication_key)

                        self.assertLessEqual(interrupted_bookmark_value[replication_key],
                                            replication_key_value,
                                            msg="Second sync bookmark was set incorrectly, a record with a greater replication-key value was synced.")

                        # Verify the second sync bookmark value is the max replication key value for a given stream
                        self.assertLessEqual(replication_key_value,
                                            second_bookmark_value[replication_key],
                                            msg="Second sync bookmark was set incorrectly, a record with a greater replication-key value was synced.")




"""
Test tap pagination of streams
"""

from math import ceil
from tap_tester import runner, connections

from base import DixaBaseTest

class DixaPaginationTest(DixaBaseTest):

    @staticmethod
    def name():
        return "tap_tester_dixa_pagination_test"

    def get_properties(self, original: bool = True):
        """Configuration properties required for the tap."""
        return_value = {
            'start_date' : self.start_date,
            'page_size' : 10
        }
        if original:
            return return_value

        return_value["start_date"] = self.start_date
        return return_value

    def test_run(self):
        page_size = int(self.get_properties().get('page_size', self.page_size))
        conn_id = connections.ensure_connection(self)

        # Checking pagination for streams with enough data
        expected_streams = ["activity_logs"]
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs = [catalog for catalog in found_catalogs
                         if catalog.get('stream_name') in expected_streams]

        self.perform_and_verify_table_and_field_selection(conn_id, test_catalogs)

        record_count_by_stream = self.run_and_verify_sync(conn_id)

        synced_records = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):
                # expected values
                expected_primary_keys = self.expected_primary_keys()

                # collect information for assertions from syncs 1 & 2 base on expected values
                record_count_sync = record_count_by_stream.get(stream, 0)
                primary_keys_list = [tuple(message.get('data').get(expected_pk)
                                           for expected_pk in expected_primary_keys[stream])
                                     for message in synced_records.get(stream).get('messages')
                                     if message.get('action') == 'upsert']

                # verify records are more than page size so multiple page is working
                # Chunk the replicated records (just primary keys) into expected pages
                pages = []
                page_count = ceil(len(primary_keys_list) / page_size)
                for page_index in range(page_count):
                    page_start = page_index * page_size
                    page_end = (page_index + 1) * page_size
                    pages.append(set(primary_keys_list[page_start:page_end]))

                # Verify by primary keys that data is unique for each page
                for current_index, current_page in enumerate(pages):
                    with self.subTest(current_page_primary_keys=current_page):

                        for other_index, other_page in enumerate(pages):
                            if current_index == other_index:
                                continue  # don't compare the page to itself

                            self.assertTrue(current_page.isdisjoint(other_page), msg=f'other_page_primary_keys={other_page}')


import re

from tap_tester import connections, menagerie
from base import DixaBaseTest


class DixaDiscoveryTest(DixaBaseTest):

    def name(self):
        return "tap_tester_discovery_test"

    def test_run(self):
        """
        Testing that discovery creates the appropriate catalog with valid metadata.

        • Verify number of actual streams discovered match expected
        • Verify the stream names discovered were what we expect
        • Verify stream names follow naming convention
          streams should only have lowercase alphas and underscores
        • Verify there is only 1 top level breadcrumb
        • Verify there are no duplicate/conflicting metadata entries.
        • Verify primary key(s) match expectations.
        • Verify replication key(s) match expectations.
        • Verify the actual replication matches our expected replication method.
        • Verify that if there is a replication key we are doing INCREMENTAL otherwise FULL.
        • Verify that primary keys and replication keys have inclusion of automatic
        • Verify that all fields have inclusion of available metadata.
        • Verify all streams have inclusion of automatic
        """
        streams_to_test = self.expected_streams()

        conn_id = connections.ensure_connection(self)
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # Verify number of actual streams discovered match expected
        found_catalog_names = {c['tap_stream_id'] for c in found_catalogs}
        actual_streams = {name for name in found_catalog_names}
        self.assertEquals(actual_streams, streams_to_test, msg="Expected streams not found in the catalog")

        # Verify stream names follow naming convention
        # streams should only have lowercase alphas and underscores
        found_catalog_names = {c['tap_stream_id'] for c in found_catalogs}
        self.assertTrue(all([re.fullmatch(r"[a-z_]+",  name) for name in found_catalog_names]),
                        msg="One or more streams don't follow standard naming")

        for stream in streams_to_test:
            with self.subTest(stream=stream):
                # verify there is only 1 top level breadcrumb
                catalog = next(iter([catalog for catalog in found_catalogs
                                     if catalog["stream_name"] == stream]))
                self.assertIsNotNone(catalog)

                # collect expected values
                expected_primary_fields = self.expected_primary_keys()[stream]
                expected_replication_fields = self.expected_replication_keys()[stream]
                expected_automatic_fields = expected_primary_fields | expected_replication_fields
                expected_replcation_method = self.expected_replication_method()[stream]

                # collect actual values
                schema_and_metadata = menagerie.get_annotated_schema(conn_id, catalog["stream_id"])
                metadata = schema_and_metadata["metadata"]
                stream_properties = [data for data in metadata if data.get("breadcrumb") == []]

                actual_fields = []
                for md_entry in metadata:
                    if md_entry['breadcrumb'] != []:
                        actual_fields.append(md_entry['breadcrumb'][1])

                actual_primary_keys = set(stream_properties[0].get("metadata", {self.PRIMARY_KEYS: []}).get(self.PRIMARY_KEYS, []))
                actual_replication_keys = set(stream_properties[0].get("metadata", {self.REPLICATION_KEYS: []}).get(self.REPLICATION_KEYS, []))
                actual_automatic_fields = set(item.get("breadcrumb", ["properties", None])[1] for item in metadata
                                              if item.get("metadata").get("inclusion") == "automatic")
                actual_replication_method = stream_properties[0].get("metadata", {self.REPLICATION_METHOD: None}).get(self.REPLICATION_METHOD)

                ##########################################################################
                ### metadata assertions
                ##########################################################################

                # verify there is only 1 top level breadcrumb in metadata
                self.assertTrue(len(stream_properties) == 1,
                                msg="There is NOT only one top level breadcrumb for {}".format(stream) + \
                                "\nstream_properties | {}".format(stream_properties))

                # Verify there are no duplicate/conflicting metadata entries.
                self.assertEqual(len(actual_fields), len(set(actual_fields)), msg = "duplicates in the metadata entries retrieved")

                # Verify primary key(s) match expectations.
                self.assertSetEqual(expected_primary_fields, actual_primary_keys)

                # Verify the actual replication matches our expected replication method.
                self.assertEquals(expected_replcation_method, actual_replication_method, msg="Replication method does not match with expectations")

                # Verify primary key(s) match expectations.
                self.assertSetEqual(expected_replication_fields, actual_replication_keys)

                # Verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
                if expected_replication_fields:
                    self.assertEqual(self.INCREMENTAL, actual_replication_method)
                else:
                    self.assertEqual(self.FULL_TABLE, actual_replication_method)

                # Verify primary key(s) match expectations.
                self.assertSetEqual(expected_automatic_fields, actual_automatic_fields)

                # verify that primary keys are given the inclusion of automatic
                self.assertTrue(actual_primary_keys.issubset(actual_automatic_fields))

                # verify that replication keys are given the inclusion of automatic
                self.assertTrue(actual_replication_keys.issubset(actual_automatic_fields))
                
                # verify that all fields have inclusion of available metadata.
                self.assertTrue(
                    all({item.get("metadata").get("inclusion") == "available"
                         for item in metadata
                         if item.get("breadcrumb", []) != []
                         and item.get("breadcrumb", ["properties", None])[1]
                         not in actual_automatic_fields}),
                    msg="Not all non key properties are set to available in metadata")
